from numpy.core.numeric import zeros_like
from SciDataTool.Functions.Plot.plot_4D import plot_4D
from SciDataTool.Functions.Plot.plot_3D import plot_3D
from SciDataTool.Functions.Plot import unit_dict, norm_dict, axes_dict
from SciDataTool.Functions.Load.import_class import import_class
from SciDataTool.Functions.parser import read_input_strings
from SciDataTool.Functions.fix_axes_order import fix_axes_order
from SciDataTool.Classes.Norm_indices import Norm_indices
from numpy import (
    any as np_any,
    where,
    meshgrid,
    unique,
    nanmax as np_max,
    nanmin as np_min,
    array2string,
    linspace,
    log10,
    array,
    ceil,
    floor,
    argmin,
)


def plot_3D_Data(
    self,
    *arg_list,
    axis_data=None,
    is_norm=False,
    unit="SI",
    save_path=None,
    x_min=None,
    x_max=None,
    y_min=None,
    y_max=None,
    z_min=None,
    z_max=None,
    z_range=None,
    is_auto_ticks=True,
    is_auto_range=True,
    is_2D_view=True,
    is_contour=False,
    is_same_size=True,
    N_stem=100,
    fig=None,
    ax=None,
    is_show_fig=None,
    is_logscale_x=False,
    is_logscale_y=False,
    is_logscale_z=False,
    thresh=None,
    is_switch_axes=False,
    colormap="RdBu_r",
    win_title=None,
    font_name="arial",
    font_size_title=12,
    font_size_label=10,
    font_size_legend=8,
    xlabel=None,
    ylabel=None,
    zlabel=None,
    title=None,
    is_disp_title=True,
    type_plot=None,
    annotation_delim=None,
    marker_color="k",
    is_shading_flat=None,
    is_hide_annotation=False,
):
    """Plots a field as a function of two axes

    Parameters
    ----------
    data : Data
        a Data object
    *arg_list : list of str
        arguments to specify which axes to plot
    is_norm : bool
        boolean indicating if the field must be normalized
    unit : str
        unit in which to plot the field
    save_path : str
        full path including folder, name and extension of the file to save if save_path is not None
    x_min : float
        minimum value for the x-axis
    x_max : float
        maximum value for the x-axis
    y_min : float
        minimum value for the y-axis
    y_max : float
        maximum value for the y-axis
    z_min : float
        minimum value for the z-axis
    z_max : float
        maximum value for the z-axis
    z_range : float
        range to use for the z-axis
    is_auto_ticks : bool
        in fft, adjust ticks to freqs (deactivate if too close)
    is_auto_range : bool
        in fft, display up to 1% of max
    is_2D_view : bool
        True to plot Data in xy plane and put z as colormap
    is_contour : bool
        True to show contour line if is_fft = False and is_2D_view = True
    is_same_size : bool
        True to have all color blocks with same size in 2D view
    N_stem : int
        number of harmonics to plot (only for stem plots)
    fig : Matplotlib.figure.Figure
        existing figure to use if None create a new one
    ax : Matplotlib.axes.Axes object
        ax on which to plot the data
    is_show_fig : bool
        True to show figure after plot
    is_logscale_x : bool
        boolean indicating if the x-axis must be set in logarithmic scale
    is_logscale_y : bool
        boolean indicating if the y-axis must be set in logarithmic scale
    is_logscale_z : bool
        boolean indicating if the z-axis must be set in logarithmic scale
    thresh : float
        threshold for automatic fft ticks
    is_switch_axes : bool
        to switch x and y axes
    is_shading_flat : bool
        True to use flat shading instead of Gouraud for pcolormesh or bilinear for pcolor
    """

    # Dynamic import to avoid import loop
    DataPattern = import_class("SciDataTool.Classes", "DataPattern")

    if len(arg_list) == 1 and type(arg_list[0]) == tuple:
        arg_list = arg_list[0]  # if called from another script with *arg_list

    axes_names = [
        axis.name for axis in read_input_strings(arg_list, axis_data=axis_data)
    ]

    # Fix axes order
    arg_list_along = fix_axes_order([axis.name for axis in self.get_axes()], arg_list)

    # Set unit
    if unit == "SI":
        unit = self.unit
    if "dB" in unit:
        if "ref" in self.normalizations:
            ref = self.normalizations["ref"].ref
        else:
            ref = 1
        unit_str = r"[" + unit + " re. " + str(ref) + " $" + self.unit + "$]"
    else:
        unit_str = r"$[" + unit + "]$"

    # Detect fft
    is_fft = False
    if any("wavenumber" in s for s in arg_list) or any("freqs" in s for s in arg_list):
        is_fft = True
        if zlabel is None:
            if self.symbol == "Magnitude":
                zlabel = "Magnitude " + unit_str
            else:
                zlabel = r"$|\widehat{" + self.symbol + "}|$ " + unit_str
    else:
        if zlabel is None:
            if is_norm:
                zlabel = (
                    r"$\frac{" + self.symbol + "}{" + self.symbol + "_0}$ " + unit_str
                )
            else:
                if self.symbol == "Magnitude":
                    zlabel = "Magnitude " + unit_str
                else:
                    zlabel = r"$" + self.symbol + "$ " + unit_str

    # Extract field and axes
    if is_fft:
        if is_2D_view:
            result = self.get_magnitude_along(
                *arg_list_along, axis_data=axis_data, unit=unit, is_norm=is_norm
            )
        else:
            result = self.get_harmonics(
                N_stem,
                *arg_list_along,
                axis_data=axis_data,
                unit=unit,
                is_norm=is_norm,
                is_flat=True,
            )
    else:
        result = self.get_along(
            *arg_list_along, axis_data=axis_data, unit=unit, is_norm=is_norm
        )

    if type_plot == "scatter" and not is_fft:
        is_fft = True
        is_same_size = False
    axes_list = result["axes_list"]
    axes_dict_other = result["axes_dict_other"]
    if axes_list[list(result.keys()).index(axes_names[0])].is_components:
        Xdata = linspace(0, len(result[axes_names[0]]) - 1, len(result[axes_names[0]]))
    else:
        Xdata = result[axes_names[0]]
    if axes_list[list(result.keys()).index(axes_names[1])].is_components:
        Ydata = linspace(0, len(result[axes_names[1]]) - 1, len(result[axes_names[1]]))
    else:
        Ydata = result[axes_names[1]]
    Zdata = result[self.symbol]
    if Xdata.size == 1:
        Zdata = Zdata[None, ...]
    if Ydata.size == 1:
        Zdata = Zdata[..., None]
    if is_fft and not is_2D_view:
        X_flat = Xdata
        Y_flat = Ydata
        Z_flat = Zdata

    else:
        Y_map, X_map = meshgrid(Ydata, Xdata)
        X_flat = X_map.flatten()
        Y_flat = Y_map.flatten()
        Z_flat = Zdata.flatten()
    if z_range is None:
        if z_min is None:
            if is_fft:
                z_min = 0
            else:
                if "dB" in unit:
                    z_min = floor(np_min(Zdata))
                else:
                    z_min = np_min(Zdata)
        if z_max is None:
            z_max = np_max(Zdata)
    else:
        if z_min is None and z_max is None:
            if "dB" in unit:
                z_max = ceil(np_max(Zdata))
            else:
                z_max = np_max(Zdata)
        if z_max is None:
            z_max = z_min + z_range
        if z_min is None:
            z_min = z_max - z_range

    # Build labels and titles
    title1 = self.name[0].capitalize() + self.name[1:] + " "
    axis = axes_list[list(result.keys()).index(axes_names[0])]
    if axis.name in axes_dict:
        name = axes_dict[axis.name]
    else:
        name = axis.name
    if axis.unit == "SI":
        axis_unit = unit_dict[axis.name]
        if xlabel is None:
            xlabel = name[0].capitalize() + name[1:] + " [" + axis_unit + "]"
    elif axis.unit in norm_dict:
        if xlabel is None:
            xlabel = norm_dict[axis.unit]
    else:
        axis_unit = axis.unit
        if xlabel is None:
            xlabel = name[0].capitalize() + name[1:] + " [" + axis_unit + "]"
    if (
        axis.name == "angle"
        and axis.unit == "°"
        and round(np_max(axis.values) / 6) % 5 == 0
    ):
        xticks = [i * round(np_max(axis.values) / 6) for i in range(7)]
    else:
        xticks = None
    if axis.is_components:
        xticklabels = result[axis.name]
        xticks = Xdata
        if annotation_delim is not None and annotation_delim in xticklabels[0]:
            annotations_x = array(
                [label.split(annotation_delim)[1] for label in xticklabels]
            )
            xticklabels = array(
                [label.split(annotation_delim)[0] for label in xticklabels]
            )
    else:
        xticklabels = None
        annotations_x = zeros_like(Xdata)

    axis = axes_list[list(result.keys()).index(axes_names[1])]
    if axis.name in axes_dict:
        name = axes_dict[axis.name]
    else:
        name = axis.name
    if axis.unit == "SI":
        axis_unit = unit_dict[axis.name]
        if ylabel is None:
            ylabel = name[0].capitalize() + name[1:] + " [" + axis_unit + "]"
    elif axis.unit in norm_dict:
        if ylabel is None:
            ylabel = norm_dict[axis.unit]
    else:
        axis_unit = axis.unit
        if ylabel is None:
            ylabel = name[0].capitalize() + name[1:] + " [" + axis_unit + "]"
    if (
        axis.name == "angle"
        and axis.unit == "°"
        and round(np_max(axis.values) / 6) % 5 == 0
    ):
        yticks = [i * round(np_max(axis.values) / 6) for i in range(7)]
    else:
        yticks = None
    if axis.is_components:
        yticklabels = result[axis.name]
        yticks = Ydata
        if annotation_delim is not None and annotation_delim in yticklabels[0]:
            annotations_y = array(
                [label.split(annotation_delim)[1] for label in yticklabels]
            )
            yticklabels = array(
                [label.split(annotation_delim)[0] for label in yticklabels]
            )
        else:
            annotations_y = zeros_like(Ydata)
    else:
        yticklabels = None
        annotations_y = zeros_like(Ydata)

    if is_fft and is_2D_view and annotation_delim is not None:
        _, annotations = meshgrid(annotations_y, annotations_x)
        annotations = annotations.flatten()
    else:
        annotations = None

    # Detect discontinuous axis (Norm_indices) to use flat shading
    if is_shading_flat is None:
        is_shading_flat = False
        flat_indices = []
        for axis in axes_list:
            if axis.unit in self.axes[axis.index].normalizations:
                if isinstance(
                    self.axes[axis.index].normalizations[axis.unit], Norm_indices
                ):
                    is_shading_flat = True
                    flat_indices.append(axes_names.index(axis.name))

    title2 = "for "
    for axis in axes_list:
        if not (
            axis.extension
            in [
                "whole",
                "interval",
                "oneperiod",
                "antiperiod",
                "smallestperiod",
                "axis_data",
                "list",
            ]
            and len(axis.values) > 1
            or (len(axis.values) == 1 and len(axes_list) == 1)
        ):
            is_display = True
        else:
            is_display = False
        if is_display:
            if axis.unit == "SI":
                axis_unit = unit_dict[axis.name]
            elif axis.unit in norm_dict:
                axis_unit = norm_dict[axis.unit]
            else:
                axis_unit = axis.unit

            if isinstance(result[axis.name], str):
                axis_str = result[axis.name]
            else:
                if isinstance(result[axis.name][0], str):
                    axis_str = result[axis.name][0]
                else:
                    if result[axis.name][0] > 10:
                        fmt = "{:.5g}"
                    else:
                        fmt = "{:.3g}"
                    axis_str = (
                        array2string(
                            result[axis.name],
                            formatter={"float_kind": fmt.format},
                        )
                        .replace(" ", ", ")
                        .replace("[", "")
                        .replace("]", "")
                        + " ["
                        + axis_unit
                        + "]"
                    )

                index = None
                if axis.is_pattern and len(axis.values) == 1:
                    for axis_obj in self.get_axes():
                        if axis_obj.name == axis.name:
                            index = axis.indices[0]

                title2 += axis.name + "=" + axis_str
                if index is not None:
                    title2 += " (slice " + str(index + 1) + "), "
                else:
                    title2 += ", "

    title3 = ""
    for axis_name in axes_dict_other:
        is_display = True
        for axis in self.axes:
            if axis.name == axis_name:
                if isinstance(axis, DataPattern) and len(axis.unique_indices) == 1:
                    is_display = False
        if is_display:
            if isinstance(axes_dict_other[axis_name][0], str):
                axis_str = axes_dict_other[axis_name][0]
            else:
                if axes_dict_other[axis_name][0] > 10:
                    fmt = "{:.5g}"
                else:
                    fmt = "{:.3g}"
                axis_str = (
                    array2string(
                        axes_dict_other[axis_name][0],
                        formatter={"float_kind": fmt.format},
                    ).replace(" ", ", ")
                    + " ["
                    + axes_dict_other[axis_name][1]
                    + "], "
                )

            title3 += axis_name + "=" + axis_str

    if title2 == "for " and title3 == "":
        title2 = ""

    if title is None:
        title = title1 + title2 + title3
        title = title.rstrip(", ")

    title = title.replace("SI", "").replace(" []", "")
    xlabel = xlabel.replace("SI", "").replace(" []", "")  # Remove dimless units
    ylabel = ylabel.replace("SI", "").replace(" []", "")  # Remove dimless units
    zlabel = zlabel.replace("SI", "").replace(" []", "")  # Remove dimless units

    if is_fft:

        if thresh is None:
            if self.normalizations is not None and "ref" in self.normalizations:
                thresh = self.normalizations["ref"].ref
            else:
                thresh = 0.02

        if "dB" in unit:
            threshold = max(10 * log10(thresh) + abs(np_max(Zdata)), 0)
        else:
            threshold = abs(thresh * np_max(Zdata))
        indices_x = np_any(where(Zdata > threshold, True, False), axis=1)
        indices_y = np_any(where(Zdata > threshold, True, False), axis=0)

        xticks = Xdata[indices_x]
        annotation_threshold = None
        if xticklabels is not None:
            xticklabels = xticklabels[indices_x]
        yticks = Ydata[indices_y]
        if yticklabels is not None:
            yticklabels = yticklabels[indices_y]
        if annotations is not None:
            annotation_threshold = threshold
        if is_auto_range:
            if len(xticks) > 1:
                if x_min is None:
                    x_min = xticks[0]
                else:
                    x_min = max(x_min, xticks[0])
                if x_max is None:
                    x_max = xticks[-1]
                else:
                    x_max = min(x_max, xticks[-1])
            else:
                if x_min is None:
                    x_min = np_min(Xdata)
                else:
                    x_min = max(x_min, np_min(Xdata))
                if x_max is None:
                    x_max = np_max(Xdata)
                else:
                    x_max = min(x_max, np_max(Xdata))
            if len(yticks) > 1:
                if y_min is None:
                    if (
                        axes_list[1].name == "wavenumber"
                        or axes_list[1].corr_name == "wavenumber"
                    ):
                        y_min = max(yticks[0], -100)
                    else:
                        y_min = 0
                else:
                    y_min = max(y_min, yticks[0])
                if y_max is None:
                    if (
                        axes_list[1].name == "wavenumber"
                        or axes_list[1].corr_name == "wavenumber"
                    ):
                        y_max = min(yticks[-1], 100)
                    else:
                        y_max = yticks[-1]
                else:
                    y_max = min(y_max, yticks[-1])
            else:
                if y_min is None:
                    y_min = max(np_min(Ydata), -100)
                else:
                    y_min = max(y_min, np_min(Ydata))
                if y_max is None:
                    y_max = min(np_max(Ydata), 100)
                else:
                    y_max = min(y_max, np_max(Ydata))
        else:
            if x_min is None:
                x_min = np_min(Xdata)
            if x_max is None:
                x_max = np_max(Xdata)
            if y_min is None:
                y_min = np_min(Ydata)
            if y_max is None:
                y_max = np_max(Ydata)

        x_min = x_min - x_max * 0.05
        x_max = x_max * 1.05
        y_min = y_min - y_max * 0.2
        y_max = y_max * 1.2

        if len(xticks) == 0 or (
            len(xticks) > 20
            and not axes_list[list(result.keys()).index(axes_names[0])].is_components
        ):
            xticks = None
        if len(yticks) == 0 or (
            len(yticks) > 20
            and not axes_list[list(result.keys()).index(axes_names[1])].is_components
        ):
            yticks = None

        if not is_auto_ticks:
            xticks = None
            yticks = None
        if is_2D_view:
            if type_plot is None:
                type_plot = "scatter"
            plot_4D(
                X_flat,
                Y_flat,
                Z_flat,
                Sdata=None,
                is_same_size=is_same_size,
                x_min=x_min,
                x_max=x_max,
                y_min=y_min,
                y_max=y_max,
                z_max=z_max,
                z_min=z_min,
                title=title,
                xticks=xticks,
                yticks=yticks,
                xticklabels=xticklabels,
                yticklabels=yticklabels,
                xlabel=xlabel,
                ylabel=ylabel,
                zlabel=zlabel,
                fig=fig,
                ax=ax,
                type_plot=type_plot,
                save_path=save_path,
                is_show_fig=is_show_fig,
                is_logscale_x=is_logscale_x,
                is_logscale_y=is_logscale_y,
                is_logscale_z=is_logscale_z,
                is_switch_axes=is_switch_axes,
                colormap=colormap,
                win_title=win_title,
                font_name=font_name,
                font_size_title=font_size_title,
                font_size_label=font_size_label,
                font_size_legend=font_size_legend,
                is_grid=True,
                is_disp_title=is_disp_title,
                annotations=annotations,
                annotation_threshold=annotation_threshold,
                marker_color=marker_color,
                is_hide_annotation=is_hide_annotation,
            )
        else:
            plot_3D(
                X_flat,
                Y_flat,
                Z_flat,
                fig=fig,
                ax=ax,
                x_min=x_min,
                x_max=x_max,
                y_min=y_min,
                y_max=y_max,
                z_min=0,
                z_max=z_max,
                title=title,
                xticks=xticks,
                yticks=yticks,
                xticklabels=xticklabels,
                yticklabels=yticklabels,
                xlabel=xlabel,
                ylabel=ylabel,
                zlabel=zlabel,
                type_plot="stem",
                save_path=save_path,
                is_show_fig=is_show_fig,
                is_logscale_x=is_logscale_x,
                is_logscale_y=is_logscale_y,
                is_logscale_z=is_logscale_z,
                is_switch_axes=is_switch_axes,
                colormap=colormap,
                win_title=win_title,
                font_name=font_name,
                font_size_title=font_size_title,
                font_size_label=font_size_label,
                font_size_legend=font_size_legend,
                is_disp_title=is_disp_title,
            )
    else:
        if x_min is None:
            x_min = np_min(Xdata)
        if x_max is None:
            x_max = np_max(Xdata)
        if y_min is None:
            y_min = np_min(Ydata)
        if y_max is None:
            y_max = np_max(Ydata)
        if is_2D_view:
            if type_plot is None:
                type_plot = "pcolor"
            if is_shading_flat and type_plot == "pcolormesh":
                # 0.5 offset
                if 0 in flat_indices:
                    Xdata = Xdata - 0.5
                    x_min -= 0.5
                    x_max -= 0.5
                if 1 in flat_indices:
                    Ydata = Ydata - 0.5
                    y_min -= 0.5
                    y_max -= 0.5
                Ydata, Xdata = meshgrid(Ydata, Xdata)
            plot_3D(
                Xdata,
                Ydata,
                Zdata,
                x_min=x_min,
                x_max=x_max,
                y_min=y_min,
                y_max=y_max,
                z_max=z_max,
                z_min=z_min,
                xlabel=xlabel,
                ylabel=ylabel,
                zlabel=zlabel,
                title=title,
                xticks=xticks,
                yticks=yticks,
                xticklabels=xticklabels,
                yticklabels=yticklabels,
                fig=fig,
                ax=ax,
                type_plot=type_plot,
                is_contour=is_contour,
                is_shading_flat=is_shading_flat,
                save_path=save_path,
                is_show_fig=is_show_fig,
                is_logscale_x=is_logscale_x,
                is_logscale_y=is_logscale_y,
                is_logscale_z=is_logscale_z,
                is_switch_axes=is_switch_axes,
                colormap=colormap,
                win_title=win_title,
                font_name=font_name,
                font_size_title=font_size_title,
                font_size_label=font_size_label,
                font_size_legend=font_size_legend,
                is_disp_title=is_disp_title,
            )
        else:
            plot_3D(
                X_map,
                Y_map,
                Zdata,
                fig=fig,
                ax=ax,
                x_min=x_min,
                x_max=x_max,
                y_min=y_min,
                y_max=y_max,
                z_min=z_min,
                z_max=z_max,
                title=title,
                xlabel=xlabel,
                ylabel=ylabel,
                zlabel=zlabel,
                yticks=yticks,
                xticklabels=xticklabels,
                yticklabels=yticklabels,
                type_plot="surf",
                save_path=save_path,
                is_show_fig=is_show_fig,
                is_logscale_x=is_logscale_x,
                is_logscale_y=is_logscale_y,
                is_logscale_z=is_logscale_z,
                is_switch_axes=is_switch_axes,
                colormap=colormap,
                win_title=win_title,
                font_name=font_name,
                font_size_title=font_size_title,
                font_size_label=font_size_label,
                font_size_legend=font_size_legend,
                is_disp_title=is_disp_title,
                is_shading_flat=is_shading_flat,
            )



from PySide2.QtWidgets import QSpacerItem, QWidget
from PySide2.QtCore import Signal

from SciDataTool.GUI.WAxisManager.Ui_WAxisManager import Ui_WAxisManager
from SciDataTool.GUI.WSliceOperator.WSliceOperator import WSliceOperator
from SciDataTool.Functions import axes_dict, rev_axes_dict
from SciDataTool.Functions.Plot import ifft_dict, fft_dict


EXTENSION_DICT = {
    "slice": ["max", "min", "rss", "sum", "rms", "mean", "list", "single"],
    "axis": [
        "derivate",
        "oneperiod",
        "antiperiod",
        "smallestperiod",
        "pattern",
        "axis_data",
        "interval",
        "whole",
    ],
}


class WAxisManager(Ui_WAxisManager, QWidget):
    """Widget that will handle the selection of the axis as well as generating WDataExtractor"""

    refreshNeeded = Signal()
    refreshRange = Signal()
    generateAnimation = Signal()

    def __init__(self, parent=None):
        """Initializing the widget by hiding/showing widget and connecting buttons

        Parameters
        ----------
        self : WAxisManager
            a WAxisManager object
        parent : QWidget
            The parent widget
        """

        # Build the interface according to the .ui file
        QWidget.__init__(self, parent=parent)
        self.setupUi(self)

        self.axes_list = list()
        self.path_to_image = None

        # Managing the signal emitted by the WAxisSelector widgets
        self.w_axis_1.axisChanged.connect(self.axis_1_updated)
        self.w_axis_2.axisChanged.connect(self.axis_2_updated)

        # The action in axis 2 is by default the one chosen in axis 1
        self.w_axis_1.actionChanged.connect(lambda: self.fft_sync("axis 1"))
        self.w_axis_2.actionChanged.connect(lambda: self.fft_sync("axis 2"))

        self.w_axis_1.refreshNeeded.connect(self.update_needed)
        self.w_axis_2.refreshNeeded.connect(self.update_needed)

    def axis_1_updated(self):
        """Method that remove the axis selected in w_axis_1 from w_axis_2 and call the method that generates
        the layout with the WDataExtractor.
        Parameters
        ----------
        self : WAxisManager
            a WAxisManager object
        """
        self.blockSignals(True)
        # Making sure that when axis 1 is updated, axis 1 and 2 are both on "None" for the action combobox
        self.fft_sync("axis 1", is_refresh=False)

        # Recovering the axis selected by the user removing it from the the second axis combobox
        self.w_axis_2.remove_axis(self.w_axis_1.get_axis_selected())
        self.gen_slice_op(is_refresh=False)
        self.blockSignals(False)
        self.update_needed()

    def axis_2_updated(self):
        """Method that make sure that when axis 2 is selected (None->?) it has the same fft/ifft combobox selected as axis1
         then generates Data Selection
        Parameters
        ----------
        self : WAxisManager
            a WAxisManager object
        """
        # Making sure that when axis 2 is updated, axis 1 and 2 are both on "None" for the action combobox
        self.blockSignals(True)
        self.fft_sync("axis 2", is_refresh=False)
        self.blockSignals(False)
        self.update_needed()

    def gen_slice_op(self, axes_request_list=None, is_refresh=True):
        """Method that gen the right WDataExtrator widget according to the axis selected by the user in the UI
        Parameters
        ----------
        self : WAxisManager
            a WAxisManager object
        """

        # Step 1 : Recovering the axis that must be generated (those that are not selected)
        # Getting all the possible axes
        axes_list_1 = self.w_axis_1.get_axes_name()[:]
        axes_list_2 = self.w_axis_2.get_axes_name()[:]

        # Getting the axes selected and removing them from the right axes_list
        axis_selected_1 = self.w_axis_1.get_axis_selected()
        axis_selected_2 = self.w_axis_2.get_axis_selected()

        if axis_selected_1 in axes_list_1:
            axes_list_1.remove(axis_selected_1)
        if axis_selected_2 in axes_list_2:
            axes_list_2.remove(axis_selected_2)

        # Selecting the axes that are in common between the two axes lists
        axes_gen = list()
        axes_gen = [ax for ax in axes_list_1 if ax in axes_list_2]

        for ax in self.axes_list:
            if ax.is_overlay:
                axes_gen.append(ax.name)

        # Step 2 : Removing the items that are in the layout currently
        for i in reversed(range(self.lay_data_extract.count())):
            widget_to_del = self.lay_data_extract.takeAt(i).widget()
            widget_to_del.deleteLater()

        # Step 3 : For each axis available, adding a WSliceOperator widget inside the layout
        # If there are no slice to do (two axis available and selected before) then we hide the groupBox
        if len(axes_gen) != 0:
            self.g_data_extract.show()
            self.w_slice_op = list()

            for axis in axes_gen:
                temp = WSliceOperator(
                    self.g_data_extract, path_to_image=self.path_to_image
                )
                temp.setObjectName(axis)
                for i, ax in enumerate(self.axes_list):
                    if (
                        ax.name == axis
                        or axis in axes_dict
                        and ax.name in axes_dict[axis]
                        or axis in rev_axes_dict
                        and ax.name in rev_axes_dict[axis]
                    ):
                        if axes_request_list is not None and axis in [
                            x.name for x in axes_request_list
                        ]:
                            temp.update(
                                ax,
                                axis_request=axes_request_list[
                                    [x.name for x in axes_request_list].index(axis)
                                ],
                            )
                        else:
                            temp.update(ax)
                temp.refreshNeeded.connect(self.update_needed)
                temp.generateAnimation.connect(self.gen_animate)
                self.w_slice_op.append(temp)
                self.lay_data_extract.addWidget(temp)

        else:
            self.w_slice_op = list()
            self.g_data_extract.hide()
        if is_refresh:
            self.update_needed()

    def gen_animate(self):
        """Methods called after clicking on animate button to generate a gif on the axis selected and display it
        Parameters
        ----------
        self : WAxisManager
            a WAxisManager object

        Output
        ---------
        None
        """

        self.generateAnimation.emit()

    def get_axes_selected(self):
        """Method that return the axes chosen by the user and their unit as a string
        so that we can use them to plot the data.
        Parameters
        ----------
        self : WAxisManager
            a WAxisManager object
        Output
        ---------
        string
            name of the axis and their units
        """

        axes_selected = list()

        # Recovering the first axis
        axes_selected.append(self.w_axis_1.get_axis_unit_selected())

        # If a second axis is selected, then we add it as well
        if self.w_axis_2.get_axis_unit_selected() != "None":
            axes_selected.append(self.w_axis_2.get_axis_unit_selected())

        return axes_selected

    def get_operation_selected(self):
        """Method that return the operations chosen by the user and the related axis as a string
         so that we can use them to plot the data.
        Parameters
        ----------
        self : WAxisManager
            a WAxisManager object
        Output
        ---------
        string
            name of the operation and its axis
        """
        operations_selected = list()

        for wid in self.w_slice_op:
            ope_selected, is_animate = wid.get_operation_selected()

            if is_animate == True:

                axis = ope_selected.split("=")[0]
                unit = "{" + ope_selected.split("{")[1]
                operations_selected.append(axis + "[oneperiod]" + unit + "to_animate")
                wid.is_animate = False
            else:
                operations_selected.append(ope_selected)

        return operations_selected

    def fft_sync(self, axis_changed, is_refresh=True):
        """Method that will check the action chosen and that update the other action combobox to have the same action.
        So that, by default, we have FFT and FFT or "None" and "None"
        Parameters
        ----------
        self : WAxisManager
            a WAxisManager object

        """
        self.blockSignals(True)
        if axis_changed == "axis 1":  # and "FFT" in [
            #     self.w_axis_1.c_action.itemText(i)
            #     for i in range(self.w_axis_1.c_action.count())
            # ]:
            action_selected = self.w_axis_1.get_current_action_name()
            self.w_axis_2.set_action(action_selected)
            if action_selected == "FFT":
                if self.w_axis_2.axis_selected in fft_dict:
                    if self.w_axis_2.axis_selected in self.w_axis_2.axes_list:
                        self.w_axis_2.axes_list[
                            self.w_axis_2.axes_list.index(self.w_axis_2.axis_selected)
                        ] = fft_dict[self.w_axis_2.axis_selected]
                        self.w_axis_2.axis_selected = fft_dict[
                            self.w_axis_2.axis_selected
                        ]
                    else:
                        self.w_axis_2.axis_selected = fft_dict[
                            self.w_axis_2.axis_selected
                        ]
                elif self.w_axis_2.axis_selected in ifft_dict:
                    if (
                        self.w_axis_2.axis_selected not in self.w_axis_2.axes_list
                        and ifft_dict[self.w_axis_2.axis_selected]
                        in self.w_axis_2.axes_list
                    ):
                        self.w_axis_2.axes_list[
                            self.w_axis_2.axes_list.index(
                                ifft_dict[self.w_axis_2.axis_selected]
                            )
                        ] = self.w_axis_2.axis_selected
            else:
                if self.w_axis_2.axis_selected in ifft_dict:
                    if self.w_axis_2.axis_selected in self.w_axis_2.axes_list:
                        self.w_axis_2.axes_list[
                            self.w_axis_2.axes_list.index(self.w_axis_2.axis_selected)
                        ] = ifft_dict[self.w_axis_2.axis_selected]
                        self.w_axis_2.axis_selected = ifft_dict[
                            self.w_axis_2.axis_selected
                        ]
                    else:
                        self.w_axis_2.axis_selected = ifft_dict[
                            self.w_axis_2.axis_selected
                        ]
                elif self.w_axis_2.axis_selected in fft_dict:
                    if (
                        self.w_axis_2.axis_selected not in self.w_axis_2.axes_list
                        and fft_dict[self.w_axis_2.axis_selected]
                        in self.w_axis_2.axes_list
                    ):
                        self.w_axis_2.axes_list[
                            self.w_axis_2.axes_list.index(
                                fft_dict[self.w_axis_2.axis_selected]
                            )
                        ] = self.w_axis_2.axis_selected
                # Make sure that all axes are reset to ifft
                for i, axis in enumerate(self.w_axis_1.axes_list):
                    if axis in ifft_dict:
                        self.w_axis_1.axes_list[i] = ifft_dict[axis]
            self.gen_slice_op(is_refresh=False)

        elif axis_changed == "axis 2":  # and "FFT" in [
            #     self.w_axis_2.c_action.itemText(i)
            #     for i in range(self.w_axis_2.c_action.count())
            # ]:
            action_selected = self.w_axis_2.get_current_action_name()
            self.w_axis_1.set_action(action_selected)
            if action_selected == "FFT":
                if self.w_axis_1.axis_selected in fft_dict:
                    if self.w_axis_1.axis_selected in self.w_axis_1.axes_list:
                        self.w_axis_1.axes_list[
                            self.w_axis_1.axes_list.index(self.w_axis_1.axis_selected)
                        ] = fft_dict[self.w_axis_1.axis_selected]
                        self.w_axis_1.axis_selected = fft_dict[
                            self.w_axis_1.axis_selected
                        ]
                    else:
                        self.w_axis_1.axis_selected = fft_dict[
                            self.w_axis_1.axis_selected
                        ]
                elif self.w_axis_1.axis_selected in ifft_dict:
                    if (
                        self.w_axis_1.axis_selected not in self.w_axis_1.axes_list
                        and ifft_dict[self.w_axis_1.axis_selected]
                        in self.w_axis_1.axes_list
                    ):
                        self.w_axis_1.axes_list[
                            self.w_axis_1.axes_list.index(
                                ifft_dict[self.w_axis_1.axis_selected]
                            )
                        ] = self.w_axis_1.axis_selected
            else:
                if self.w_axis_1.axis_selected in ifft_dict:
                    if self.w_axis_1.axis_selected in self.w_axis_1.axes_list:
                        self.w_axis_1.axes_list[
                            self.w_axis_1.axes_list.index(self.w_axis_1.axis_selected)
                        ] = ifft_dict[self.w_axis_1.axis_selected]
                        self.w_axis_1.axis_selected = ifft_dict[
                            self.w_axis_1.axis_selected
                        ]
                    else:
                        self.w_axis_1.axis_selected = ifft_dict[
                            self.w_axis_1.axis_selected
                        ]
                elif self.w_axis_1.axis_selected in fft_dict:
                    if (
                        self.w_axis_1.axis_selected not in self.w_axis_1.axes_list
                        and fft_dict[self.w_axis_1.axis_selected]
                        in self.w_axis_1.axes_list
                    ):
                        self.w_axis_1.axes_list[
                            self.w_axis_1.axes_list.index(
                                fft_dict[self.w_axis_1.axis_selected]
                            )
                        ] = self.w_axis_1.axis_selected
                # Make sure that all axes are reset to ifft
                for i, axis in enumerate(self.w_axis_2.axes_list):
                    if axis in ifft_dict:
                        self.w_axis_2.axes_list[i] = ifft_dict[axis]
            self.gen_slice_op(is_refresh=False)
        self.w_axis_1.set_unit()
        self.w_axis_2.set_unit()
        self.blockSignals(False)
        if is_refresh:
            self.update_needed()

    def set_axis_widgets(
        self,
        data,
        axes_request_list,
        frozen_type=0,
        is_keep_config=False,
        path_to_image=None,
    ):
        """Method used to set the axes of the Axes group box as well as setting the widgets of the DataSelection groupbox
        Parameters
        ----------
        self : WAxisManager
            a WAxisManager object
        data : DataND
            The DataND object that we want to plot
        axes_request_list:
            list of RequestedAxis which are the info given for the autoplot (for the axes and DataSelection)
        frozen_type : int
            0 to let the user modify the axis of the plot, 1 to let him switch them, 2 to not let him change them,
            3 to freeze both axes and operations, 4 to freeze fft, 5 to only allow switch
        path_to_image : str
            path to the folder where the image for the animation button is saved
        """
        # Setting path to recover the image for the animate button
        self.path_to_image = path_to_image
        self.w_axis_1.path_to_image = self.path_to_image
        self.w_axis_2.path_to_image = self.path_to_image

        if is_keep_config:  # Only update slider
            for wid in self.w_slice_op:
                if hasattr(wid, "axis_value"):
                    wid.update_floatEdit()
            axes_list = data.get_axes()

        else:
            # Step 1 : If only one axis is given with the object, then we hide w_axis_2 and g_data_extract
            # We also have to hide if we have more than one axis but one have is_overlay = True
            if len(data.get_axes()) == 1 or (
                len(data.get_axes())
                - len([ax for ax in data.get_axes() if ax.is_overlay == True])
                == 1
            ):
                self.w_axis_2.hide()
                self.g_data_extract.hide()

            else:
                self.w_axis_2.show()
                self.g_data_extract.show()

            # Step 2 : If we have user input, the we set the UI according to user_input.
            # Otherwise we use the default info
            self.w_axis_1.blockSignals(True)
            self.w_axis_2.blockSignals(True)

            # Reinitialize filter indices
            self.w_axis_1.indices = None
            self.w_axis_2.indices = None

            if axes_request_list == []:
                # Case where no user_input was given
                # Sending the info of data to the widget (mainly the axis)
                self.axes_list = data.get_axes()
                self.w_axis_1.update(self.axes_list)
                self.w_axis_2.update(self.axes_list, axis_name="Y")

                # Updating w_axis_2 according to w_axis_1 then generating DataSelection
                self.axis_1_updated()
                axes_list = self.axes_list

            else:
                # Case where a userinput was given (auto plot)

                # If user_input are given (auto-plot), we have to process them
                axes_list = [
                    ax
                    for ax in axes_request_list
                    if ax.extension in EXTENSION_DICT["axis"]
                ]

                slices_op_list = [
                    ax
                    for ax in axes_request_list
                    if ax.extension in EXTENSION_DICT["slice"]
                ]

                # Sending the info of data to the widget (mainly the axis)
                self.axes_list = data.get_axes()
                self.w_axis_1.update(self.axes_list)
                self.w_axis_2.update(self.axes_list, axis_name="Y")
                # Setting the axis selected in w_axis_1 according to user_input_list
                self.w_axis_1.set_axis(axes_list[0])

                # Updating w_axis_2 according to w_axis_1 then generating DataSelection
                self.axis_1_updated()

                # Setting the axis selected in w_axis_1 according to user_input_list if we have a second axis
                if len(axes_list) == 2:
                    self.w_axis_2.set_axis(axes_list[1])

                # Making sure that we have the same fft/ifft selected for both axis
                # self.axis_2_updated()

                # Generating DataSelection with the input of user if they are given or by default (like in a manual plot)
                if len(slices_op_list) != 0:
                    self.gen_slice_op(axes_request_list)
                    self.set_slice_op(slices_op_list)
                else:
                    self.gen_slice_op(axes_request_list)

        # Depending on the value of frozen type we are going to act on the UI
        if frozen_type == 1 and len(axes_list) == 2:
            # Recovering the axis requested by the user as they will be soft frozen (possible to switch but not possible to choose another axis)
            axes_list_name = [ax.name for ax in axes_list]
            axes_soft_frozen = [
                ax for ax in self.axes_list if ax.name in axes_list_name
            ]

            # We update the WAxisSelector widget with the axes that will be soft frozen
            self.w_axis_1.update(axes_soft_frozen)
            self.w_axis_2.update(axes_soft_frozen)

            self.axis_1_updated()

        elif frozen_type == 2:
            # If we want to hard freeze the axis, we just have to disable the axis comboboxes
            self.w_axis_1.c_axis.setDisabled(True)
            self.w_axis_2.c_axis.setDisabled(True)

        elif frozen_type == 3:
            # Freezing the axes
            self.w_axis_1.c_axis.setDisabled(True)
            self.w_axis_1.c_action.setDisabled(True)
            self.w_axis_2.c_axis.setDisabled(True)
            self.w_axis_2.c_action.setDisabled(True)

            # Freezing the operations (allowing only slice and requested ope)
            for w_slice in self.w_slice_op:
                current_ope = w_slice.c_operation.currentText()
                if current_ope != "slice":
                    w_slice.c_operation.blockSignals(True)
                    w_slice.c_operation.clear()
                    w_slice.c_operation.addItems(["slice", current_ope])
                    w_slice.c_operation.setCurrentIndex(1)
                    w_slice.c_operation.blockSignals(False)
                else:
                    w_slice.c_operation.setDisabled(True)
                # w_slice.lf_value.setDisabled(True)
                # w_slice.slider.setDisabled(True)

        elif frozen_type == 4:
            self.w_axis_1.c_axis.setDisabled(True)
            self.w_axis_2.c_axis.setDisabled(True)
            if axes_list[0].name in ifft_dict or axes_list[0].name in fft_dict:
                self.w_axis_1.c_action.setDisabled(True)
            if len(axes_list) == 2 and (
                axes_list[1].name in ifft_dict or axes_list[0].name in fft_dict
            ):
                self.w_axis_2.c_action.setDisabled(True)

        elif frozen_type == 5:
            # Freezing the axes
            self.w_axis_1.c_action.setDisabled(True)
            self.w_axis_2.c_action.setDisabled(True)

            # Freezing the operations
            for w_slice in self.w_slice_op:
                w_slice.b_action.setDisabled(True)
                w_slice.c_operation.setDisabled(True)
                w_slice.lf_value.setDisabled(True)
                w_slice.slider.setDisabled(True)

        if len(axes_list) == 1:
            self.w_axis_2.hide()

        self.w_axis_1.blockSignals(False)
        self.w_axis_2.blockSignals(False)

    def set_slice_op(self, user_input_list):
        """Method that set the right operation inside each WSliceOperator inside of w_slice_op
        according to user input (auto plot).
        Parameters
        ----------
        self : WAxisManager
            a WAxisManager object
        user_input_list : list
            list of the inputs from the user to set the DataSelection (auto-plot)
        """
        for wid in self.w_slice_op:
            if self.w_slice_op.index(wid) < len(user_input_list):
                wid.set_operation(user_input_list[self.w_slice_op.index(wid)])

    def update_needed(self):
        """Method that emits a signal (refreshNeeded) that will be used to automaticaly update the plot inside the GUI.
        This signal is triggered by other signals comming from WSliceOperator or WAxisSelector.
        refreshRange is a different signal that we use to update the values of min and max inside w_range
        Parameters
        ----------
        self : WAxisManager
            a WAxisManager object
        """

        self.refreshRange.emit()
        self.refreshNeeded.emit()



from matplotlib import use

# Set Matplotlib backend
use("Qt5Agg")  # Use PySide2 backend


def update_cb_enable(combobox):
    # To disable a combobox with only one item
    if combobox.count() <= 1:
        combobox.setEnabled(False)
    else:
        combobox.setEnabled(True)


from os.path import dirname, abspath, join

DATA_DIR = join(abspath(dirname(__file__)), "Results")



from SciDataTool.GUI.DDataPlotter.DDataPlotter import DDataPlotter
from sys import argv, exit
from PySide2.QtWidgets import QApplication
from PySide2 import QtCore
from SciDataTool.Functions import parser

QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)


def plot(
    self,
    *args,
    component=None,
    unit=None,
    z_min=None,
    z_max=None,
    is_auto_refresh=False,
    is_show_fig=True,
    is_create_appli=True,
    frozen_type=0,
):
    """Plot the Data object in the GUI

    Parameters:
    -----------
    self : VectorField
        A VectorField object
    *args : 1 or 2 str
        Example ("time", "angle[0]") or ("angle")
    component : str
        Name of the component to plot (For VectorField only)
    unit : str
        unit in which to plot the field
    z_min : float
        Minimum value for Z axis (or Y if only one axe)
    z_max : float
        Minimum value for Z axis (or Y if only one axe)
    is_auto_refresh : bool
        True to refresh at each widget changed (else wait call to button)
    is_show_fig : bool
        To show the GUI or return the widget (False for testing)
    is_create_appli : bool
        True to create an QApplication (required if not already created by another GUI)
    frozen_type : int
        0 to let the user modify the axis of the plot, 1 to let him switch them, 2 to not let him change them, 3 to freeze both axes and operations, 4 to freeze fft
    """

    if is_create_appli:
        a = QApplication(argv)

    # Parse the first arguments to get the axes
    axes_request_list = parser.read_input_strings(
        [arg for arg in args if arg != None], axis_data=None
    )

    if unit is None:
        unit = self.components[list(self.components.keys())[0]].unit

    wid = DDataPlotter(
        data=self,
        axes_request_list=axes_request_list,
        component=component,
        unit=unit,
        z_min=z_min,
        z_max=z_max,
        is_auto_refresh=is_auto_refresh,
        frozen_type=frozen_type,
    )

    if is_show_fig:
        wid.show()

    if is_create_appli:
        exit(a.exec_())
    else:
        return wid



from SciDataTool.Methods.DataDual import is_freq


def get_along(
    self,
    *args,
    unit="SI",
    is_norm=False,
    axis_data=[],
    is_squeeze=True,
    is_magnitude=False,
    corr_unit=None,
):
    """Returns the ndarray of the field, using conversions and symmetries if needed.
    Parameters
    ----------
    self: Data
        a Data object
    *args: list of strings
        List of axes requested by the user, their units and values (optional)
    unit: str
        Unit requested by the user ("SI" by default)
    is_norm: bool
        Boolean indicating if the field must be normalized (False by default)
    axis_data: list
        list of ndarray corresponding to user-input data
    Returns
    -------
    list of 1Darray of axes values, ndarray of field values
    """

    # Get best attributes (time/space or fourier)
    if is_freq(*args):
        if hasattr(self, "values_df") and self.values_df is not None:
            # Replace axes and values with fourier ones
            self.axes = self.axes_df
            self.values = self.values_df
        else:
            # Replace axes and values with time/space ones
            self.axes = self.axes_dt
            self.values = self.values_dt
    else:
        if hasattr(self, "values_dt") and self.values_dt is not None:
            # Replace axes and values with time/space ones
            self.axes = self.axes_dt
            self.values = self.values_dt
        else:
            # Replace axes and values with fourier ones
            self.axes = self.axes_df
            self.values = self.values_df

    # Call DataND.get_along()
    return super(type(self), self).get_along(
        *args,
        unit=unit,
        is_norm=is_norm,
        axis_data=axis_data,
        is_squeeze=is_squeeze,
        is_magnitude=is_magnitude,
        corr_unit=corr_unit,
    )
